import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

import 'anchored_popups.dart';

enum PopUpMode {
  clickToToggle, // Click a region to open PopOver, click barrier to close.
  hover, // Open on hoverIn (slightly delayed), close on hoverOut
}

typedef Widget PopupButtonBuilder(BuildContext context, Widget child, VoidCallback show);

class AnchoredPopUpRegion extends StatefulWidget {
  AnchoredPopUpRegion(
      {Key? key,
      required this.child,
      required this.popChild,
      this.anchor,
      this.popAnchor,
      this.barrierDismissable,
      this.barrierColor,
      this.mode = PopUpMode.hover,
      this.delay,
      this.buttonBuilder})
      : super(key: key);
  final Widget child;
  final Widget popChild;
  final Alignment? anchor;
  final Alignment? popAnchor;
  final bool? barrierDismissable;
  final Color? barrierColor;
  final PopUpMode mode;
  final Duration? delay;
  final PopupButtonBuilder? buttonBuilder;
  @override
  AnchoredPopUpRegionState createState() => AnchoredPopUpRegionState();

  // Non-interactive tool-tips, triggered on a delayed hover. Auto-close when you roll-out of the PopOverRegion
  static AnchoredPopUpRegion hover(
      {Key? key,
      required Widget child,
      required Widget popChild,
      Alignment? anchor,
      Alignment? popAnchor,
      Duration? delay}) {
    return AnchoredPopUpRegion(
        key: key,
        child: child,
        popChild: popChild,
        anchor: anchor,
        popAnchor: popAnchor,
        mode: PopUpMode.hover,
        delay: delay);
  }

  // Click to open/close. Use for interactive panels, or other elements that should close themselves
  static AnchoredPopUpRegion click(
      {Key? key,
      required Widget child,
      required Widget popChild,
      Alignment? anchor,
      Alignment? popAnchor,
      bool? barrierDismissable,
      Color? barrierColor,
      PopupButtonBuilder? buttonBuilder}) {
    return AnchoredPopUpRegion(
      key: key,
      child: child,
      popChild: popChild,
      anchor: anchor,
      popAnchor: popAnchor,
      mode: PopUpMode.clickToToggle,
      barrierColor: barrierColor,
      barrierDismissable: barrierDismissable,
      buttonBuilder: buttonBuilder,
    );
  }

  static AnchoredPopUpRegion hoverWithClick({
    Key? key,
    required Widget child,
    required Widget hoverPopChild,
    required Widget clickPopChild,
    bool barrierDismissable = true,
    Color? barrierColor,
    Alignment? hoverAnchor,
    Alignment? hoverPopAnchor,
    Alignment? clickAnchor,
    Alignment? clickPopAnchor,
    PopupButtonBuilder? buttonBuilder,
  }) {
    return click(
        key: key,
        anchor: clickAnchor,
        barrierColor: barrierColor,
        barrierDismissable: barrierDismissable,
        popChild: clickPopChild,
        popAnchor: clickPopAnchor,
        buttonBuilder: buttonBuilder,
        child: hover(popAnchor: hoverPopAnchor, popChild: hoverPopChild, anchor: hoverAnchor, child: child));
  }
}

// TODO: Support for some sort of dualMode widget would be nice, has both a tooltip and a panel.
class AnchoredPopUpRegionState extends State<AnchoredPopUpRegion> {
  Timer? _timer;
  AnchoredPopupsController? _popups;

  @override
  Widget build(BuildContext context) {
    Widget content;
    // If Hover, add a MouseRegion
    if (widget.mode == PopUpMode.hover) {
      content = MouseRegion(
        opaque: true,
        onEnter: (_) => _handleHoverStart(),
        onExit: (_) => _handleHoverEnd(),
        child: widget.child,
      );
    } else {
      content = widget.buttonBuilder?.call(context, widget.child, _show) ??
          MouseRegion(cursor: SystemMouseCursors.click, child: GestureDetector(onTap: _show, child: widget.child));
    }
    return content;
  }

  void cancelTimer() => _timer?.cancel();

  @override
  void dispose() {
    if (widget.mode == PopUpMode.hover) {
      _handleHoverEnd();
    }
    super.dispose();
  }

  void _show() {
    if (mounted == false) {
      return;
    }
    _popups = AnchoredPopups.of(context);
    _popups?.show(context,
        popUpMode: widget.mode,
        popContent: widget.popChild,
        anchor: widget.anchor ?? Alignment.bottomCenter,
        popAnchor: widget.popAnchor ?? Alignment.topCenter,
        useBarrier: widget.mode != PopUpMode.hover,
        barrierColor: widget.barrierColor ?? Colors.transparent,
        dismissOnBarrierClick: widget.barrierDismissable ?? true);
  }

  void _handleHoverStart() {
    _timer?.cancel();
    _timer = Timer.periodic(widget.delay ?? Duration(milliseconds: 400), (_) {
      _show();
      _timer?.cancel();
    });
  }

  void _handleHoverEnd() {
    cancelTimer();
    bool isStillOpen = _popups?.currentPopup?.context == context;
    if (isStillOpen) _popups?.hide();
  }
}
